  # ****************************************************************************
  #
  # This file is part of the yasmine editing tool.
  #
  # yasmine (Yet Another Station Metadata INformation Editor), a tool to
  # create and edit station metadata information in FDSN stationXML format,
  # is a common development of IRIS and RESIF.
  # Development and addition of new features is shared and agreed between * IRIS and RESIF.
  #
  #
  # Version 1.0 of the software was funded by SAGE, a major facility fully
  # funded by the National Science Foundation (EAR-1261681-SAGE),
  # development done by ISTI and led by IRIS Data Services.
  # Version 2.0 of the software was funded by CNRS and development led by * RESIF.
  #
  # This program is free software; you can redistribute it
  # and/or modify it under the terms of the GNU Lesser General Public
  # License as published by the Free Software Foundation; either
  # version 3 of the License, or (at your option) any later version. *
  # This program is distributed in the hope that it will be
  # useful, but WITHOUT ANY WARRANTY; without even the implied warranty
  # of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  # GNU Lesser General Public License (GNU-LGPL) for more details. *
  # You should have received a copy of the GNU Lesser General Public
  # License along with this software. If not, see
  # <https://www.gnu.org/licenses/>
  #
  #
  # 2019/10/07 : version 2.0.0 initial commit
  #
  # ****************************************************************************/

from sys import exit
import argparse
import os
import sys
import yaml

from .. import installation_dir, yml_template_dir

import logging
logger = logging.getLogger()

from .libs_obs import read_yml_file, show_fields, check_field
from .libs_log import string_to_logLevel

list_fields = {'comments', 'equipments', 'identifiers', 'operators', 'types', 'external_references'}
root_fields = {'source', 'sender', 'module', 'module_uri'}

TEMPLATE_DIR = yml_template_dir()

def read_config(requireConfigFile=False):
    '''
    Read yaml configfil
    '''

    config = None

    # From highest to lowest priority location of config.yml:
    # 1. passed on cmd line --configFile=...
    # 2. os.getcwd() = Location script is called from
    # 3. INSTALL_DIR = Location of yasmine_cli directory
    # 4. None

    configfile = None
    # See if it was passed on cmd line:
    configfile = pick_off_configFile(sys.argv)

    if configfile is None:
        # Look for config.yml in dir where pdl-to-aqms is called from (changes)
        test_path = os.path.join(os.getcwd(), 'config.yml')
        if os.path.exists(test_path):
            configfile = test_path
    if configfile is None:
        # Look for config.yml in dir where pdl_to_aqms.py is located (it never changes)
        #SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))
        #test_path = os.path.join(SCRIPT_DIR, 'config.yml')
        INSTALL_DIR = installation_dir()
        test_path = os.path.join(INSTALL_DIR, 'config.yml')
        if os.path.exists(test_path):
            configfile = test_path

    if configfile:
        config = configure(configfile)
    else:
        if requireConfigFile:
            logger.ERROR("ConfigFile is required but none has been found --> Exit")
            exit(2)

    return config

def pick_off_configFile(argv):
    '''
    See if configFile was passed on cmd line
    '''
    configFile= None
    for word in argv:
        if '--configFile' in word:
            foo, configFile = word.split('--configFile=')
    return configFile

def processCmdLine(fname):

    epilog='''
Examples:
>python yasmine-cli.py --level_network=II --field=description --value='Network description' --infiles=...
>python yasmine-cli.py --level_station=II.* --field=latitude --value=34.97 --infiles=...
>python yasmine-cli.py --level_channel=II.ANMO.00.* --field=comments[0] --value=yml:/path/comment.yml --infiles=...
>python yasmine-cli.py --level_network=* --action=add --from_yml=path/to/station.yml --infiles=...
    '''

    parser = argparse.ArgumentParser(
        description='',
        #usage='use "%(prog)s --help" for more information',
        epilog=''
    )

    agroup = parser.add_argument_group('level options: [default=root is implied if no level set]')
    group = agroup.add_mutually_exclusive_group()
    group.add_argument("--level_network", type=str, metavar='II')
    group.add_argument("--level_station", type=str, metavar='II.*')
    group.add_argument("--level_channel", type=str, metavar='II.ANMO.00.*')

    bgroup = parser.add_argument_group('action options: [default=update if no action set]')
    bgroup.add_argument("--action", type=str, choices=['add','delete','update','select'],
                          metavar='[add/delete/update basenode]')

    cgroup = parser.add_argument_group('epoch options: Use to filter down to epoch level')
    cgroup.add_argument("--epoch_station", type=int, metavar='int', help="station epoch index to filter on, eg, --epoch_station=1")
    cgroup.add_argument("--epoch_channel", type=int, metavar='int', help="channel epoch index to filter on, eg, --epoch_channel=0")

    dgroup = parser.add_argument_group('build options')
    dgroup.add_argument("--field", type=str, help='field, key or attribute to update. eg, --field=Latitude or --field=comments[1]')
    group = dgroup.add_mutually_exclusive_group()
    group.add_argument("--value", type=str, help='value of field, key or attribute to update. Ex. --value=34.97 or --value=yml:/path/comment.yml')
    group.add_argument("--from_yml", type=str, metavar='fname.yml', help='Used to add new basenode object created from fname.yml. eg, --from_yml=/some/path/network.yml')

    optional = parser.add_argument_group('other optional arguments')
    optional.add_argument("--infiles", type=list_str, required=False, metavar='', help='comma separated list of input xml files [default=stdin]')
    optional.add_argument('-o', '--output', type=str, metavar='', help='Name of output xml file [default=stdout]. eg, --output=foo.xml')
    optional.add_argument('-p', '--print-epochs', help='Print out sorted Station/Channel epochs', action="store_true")

    optional.add_argument('--print_all', help='Print out sorted Station/Channel epochs + operator/comment lists', action="store_true")
    optional.add_argument('--dont_validate', help='Turn OFF StationXML validation on all inputs/outputs', action="store_true")
    optional.add_argument("--schema_version", type=str, choices=['1.0', '1.1'], metavar='ver', help='{1.0, 1.1}')
    optional.add_argument('--show_fields', help='Print out allowable --field, + --value combinations', action="store_true")
    optional.add_argument('--show-fields', help='Print out allowable --field, + --value combinations', action="store_true")
    #optional.add_argument('-s', '--select', help='Select only --level_station stations', action="store_true")
    optional.add_argument('--plot_resp', help='Plot all channel responses', action="store_true")
    optional.add_argument('--plot_dir', type=str, metavar='path', help='Path to dir to save plot responses')
    optional.add_argument('--loglevel', type=str, metavar='log level', help='loglevel in {DEBUG, INFO, WARN, etc}')

    # Intercept the help msg so we can also print examples after
    if len(sys.argv) == 1 or \
       (len(sys.argv) == 2 and (sys.argv[1] == '-h' or sys.argv[1] == '--help')):
        parser.print_help()
        print(epilog)
        exit()

    args, unknown = parser.parse_known_args()

    if unknown:
        logger.error("The following cmd line params are unknown:%s" %(" ".join(unknown)))
        parser.print_usage()
        exit(2)


    if args.show_fields or getattr(args, 'show-fields', None):
        show_fields()
        exit()

    if args.loglevel:
        logger.setLevel(string_to_logLevel(args.loglevel))


# If no action set and we're not simpling printing out,
#  the action is either 'update' or 'select':
    if not args.action and not args.print_epochs and not args.print_all:
        args.action = 'update'
        if not args.field and not args.value:
            args.action = 'select'

    if args.action == 'select' and args.field:
        msg = ("You're using --action=select with --field: If you want to update the field, "
               "don't use --action (default action=update). If you want to select (=filter on a station), "
               "don't use --field.")
        logger.warning(msg)
        exit(2)

    update_pair = None
    args.use_index = False

    args.update_root = False
    if args.field in root_fields:
        args.update_root = True

    if args.action == 'update':
        args.update = 'set' # default

        if args.field and args.value:
            if not args.level_network and not args.level_station and \
               not args.level_channel and not args.update_root:
                logger.error("--field=%s is *not* a recognized root field and no --level was specified" % args.field)
                logger.error("Use yasmine-cli.py --show_fields to see all known fields -or-")
                logger.error("    yasmine-cli.py -h to see all options")
                logger.error("You must specify level where this (field,value) should be updated!")
                logger.error("Example: --level_network=II")
                logger.error("Example: --level_channel=II.*.00.*")
                #parser.print_help()
                exit(2)

            field = args.field
            # Replace --field=comments[16] --> field='comments', index=16
            use_index = False
            field_index = None
            if '[' in field and field[-1] == ']':
                try:
                    i = field.index('[') + 1
                    field_index = int(field[i:-1])
                    field = field[0:i-1]
                    use_index = True
                    args.field_index = field_index
                    args.use_index = use_index
                except ValueError as err:
                    raise
                    #logger.info("Keep it a string!")
            #logger.info("field:%s value:%s type(%s)" % (args.field, value, type(value)))


            # --value='some_string'
            # --value=['a','list','of','strings']
            # --value=[]  // empty list
            # --value=None  // remove attribute
            # --value=71.12
            # --value=[72.5, 32.7] // list of floats
            # --value=from_yml --from_yml=some_path_to_yml   // yml file specifies value
            value = args.value
            #if value == 'from_yml':
            if value[0:4] == 'yml:':  # Read/build value from yaml file
                yml_file = value[4:]
                logger.info("Read value(s) from yml file:%s" % yml_file)

                # First look for file in dir we're calling this script *from*
                file_1 = os.path.join(os.getcwd(), yml_file)
                # Else look for file in TEMPLATE_DIR
                file_2 = os.path.join(TEMPLATE_DIR, os.path.basename(yml_file))

                logger.debug("Try to read from file_1=[%s]" % file_1)
                logger.debug("Try to read from file_2=[%s]" % file_2)
                if os.path.exists(file_1):
                    ymlfile = file_1
                    logger.info("Found yml_file=[%s]" % ymlfile)
                elif os.path.exists(file_2):
                    ymlfile = file_2
                    logger.info("Found yml_file=[%s]" % ymlfile)
                else:
                    logger.error("%s: Unable to find yml_file=[%s] file in either: cwd=%s -or: TEMPLATE_DIR=%s" %
                                 (fname, yml_file, os.getcwd(), TEMPLATE_DIR))
                    exit(2)

                value = read_yml_file(ymlfile)

            elif value[0] == '[' and value[-1] == ']': # value is a list of some sort
                    if len(value) <= 3: # --value=[] or --value=[ ]  // Set to empty list (to nullify list attribs)
                        value = []
                    else:
                        new_value = []
                        strings = value[1:-1].split(',')
                        logger.info("ProcessCmdLine: --value=%s = list.  Process as list:%s" % (value, strings))
                        try:
                            val = int(strings[0])
                            logger.info("ProcessCmdLine: --value=%s = list.  Appears to be list of ints" % value)
                            for string in strings:
                                new_value.append(int(string))
                        except ValueError as err:
                            try:
                                val = float(strings[0])
                                logger.info("ProcessCmdLine: --value=%s = list.  Appears to be list of floats" % value)
                                for string in strings:
                                    new_value.append(float(string))
                            except ValueError as err:
                                try:
                                    val = str(strings[0])
                                    logger.info("ProcessCmdLine: --value=%s = list.  Appears to be list of strings" % value)
                                    for string in strings:
                                        new_value.append(str(string))
                                except ValueError as err:
                                    logger.error("I give up: Unable to determine type of list elements in --value")
                                    exit(2)
                        value = new_value

            else: # value is scalar:
                try:
                    value = int(args.value)
                    logger.debug("ProcessCmdLine: --value=%s type=int" % value)
                except ValueError as err:
                    try:
                        value = float(args.value)
                        logger.debug("ProcessCmdLine: --value=%s type=float" % value)
                    except ValueError as err:
                        logger.debug("ProcessCmdLine: --value=%s type=str" % value)
                        if value == 'None':
                            value = None
                            logger.debug("ProcessCmdLine: convert --value=None --> remove attribute")

            args.field = field
            args.value = value
            args.update_pair = (field, value)

        else:
            logger.error("--action=update requires --field=... and --value=.. (or --value=from_yml + --from_yml=path..yml")
            logger.error("Example: --action=update --field=latitude --value=71.34 --level_station=II.ANMO   // set lat on II.ANMO")
            parser.print_help()
            exit(2)

        if field in list_fields:
            if isinstance(value, list):
                if use_index:
                    logger.error("%s[%d] cannot be set to *list* value=%s (type:%s)" % \
                                (field, field_index, value, type(value)))
                    exit(2)
                else:
                    if not value:
                        logger.info("field:%s set list to emtpy list" % field)
                    else:
                        logger.info("field:%s set list --> list" % field)
            else:
                if use_index:
                    logger.info("%s is list --> set %s[%d]=value (type:%s)" % \
                                (field, field, field_index, type(value)))
                    args.update = 'list_modify'
                else:
                    logger.info("%s is list --> append value (type:%s)" % \
                                (field, type(value)))
                    args.update = 'list_append'

        level = None
        if not args.update_root:
            if args.level_network:
                level = 'network'
            elif args.level_station:
                level = 'station'
            elif args.level_channel:
                level = 'channel'
            else:
                logger.error("Unknown state: No level set for action=update!")
                exit(2)

            if not check_field(field, value, level):
                #logger.error("field=[%s] or value=[%s] does not exist for level=[%s]" % \
                            #(field, value, level))
                logger.error("Element [%s] has no field/attrib=[%s]" % \
                            (level.capitalize(), field))
                logger.info("To see what fields are allowable do: >python yasmine-cli.py --show_fields")
                exit(2)

    elif args.action == 'add':
        if not args.from_yml:
            parser.print_help()
            logger.error("--action=add is used only to insert a BaseNodeType (Network/Station/Channel)")
            logger.error("Example: --action=add --from_yml=yml/network.yml")
            logger.error("Example: --action=add --from_yml=yml/station.yml --level_network=II")
            logger.error("Example: --action=add --from_yml=yml/channel.yml --level_network=II.ANMO")
            exit(2)
        else:
            if args.from_yml[0:4] == 'yml:':  # Read/build value from yaml file
                yml_file = args.from_yml[4:]
                logger.info("Read value(s) from yml file:%s" % yml_file)

                # First look for file in dir we're calling this script *from*
                file_1 = os.path.join(os.getcwd(), yml_file)
                # Else look for file in TEMPLATE_DIR
                file_2 = os.path.join(TEMPLATE_DIR, os.path.basename(yml_file))

                logger.debug("Try to read from file_1=[%s]" % file_1)
                logger.debug("Try to read from file_2=[%s]" % file_2)
                if os.path.exists(file_1):
                    ymlfile = file_1
                    logger.info("Found yml_file=[%s]" % ymlfile)
                elif os.path.exists(file_2):
                    ymlfile = file_2
                    logger.info("Found yml_file=[%s]" % ymlfile)
                else:
                    logger.error("%s: Unable to find yml_file=[%s] file in either: cwd=%s -or: TEMPLATE_DIR=%s" %
                                 (fname, yml_file, os.getcwd(), TEMPLATE_DIR))
            else:
                ymlfile = args.from_yml

            value = read_yml_file(ymlfile)
            field = value.__class__.__name__
            args.field = field
            args.value = value
            args.update_pair = (field, value)
    elif args.action == 'delete':
        logger.info("Use delete to delete a basenode")

    scnl_filter = struct(NET=None, STA=None, CHA=None, LOC=None, STN_EPOCH=None, CHN_EPOCH=None, INDEX=None)

    if args.use_index:
        scnl_filter.INDEX = args.field_index

    level = None

    if args.level_network:
        level = 'network'
        if args.level_network == '*':
            net = None
        elif args.level_network.isalnum() and len(args.level_network) > 0 :
            net = args.level_network
        else:
            parser.print_usage()
            logger.info("Example: --level_network=IU")
            exit(2)
        scnl_filter.NET = net
    elif args.level_station:
        level = 'station'
        fail = False
        if args.level_station.isalnum():
            sta = args.level_station
            net = None
        else:
            try:
                (tmp_net, tmp_sta) = args.level_station.split(".")

                if tmp_net.isalnum():
                    net = tmp_net
                elif tmp_net == '*':
                    net = None
                else:
                    fail = True

                if tmp_sta.isalnum():
                    sta = tmp_sta
                elif tmp_sta == '*':
                    sta = None
                else:
                    fail = True
            except:
                logger.error("Unable to split net.sta")
                logger.info("Example: --level_station=ANMO    // Act on station(s)=ANMO of net=*")
                logger.info("Example: --level_station=*.ANMO  // Act on station(s)=ANMO of net=*")
                logger.info("Example: --level_station=IU.ANMO // Act on station(s)=ANMO of net=IU")
                logger.info("Example: --level_station=IU.*    // Act on station(s)=*    of net=IU")
                parser.print_usage()
                exit(2)

        scnl_filter.NET = net
        scnl_filter.STA = sta

    elif args.level_channel:
        level = 'channel'
        fail = False
        try:
            (fields) = args.level_channel.split(".")
        except:
            #print("Unable to split net.sta")
            fail = True

        if len(fields) != 4:
            fail = True
        else:
            for field in fields:
                if not field.isalnum() and field != '*' and field != '--':

                    logger.info("Example: --level_channel=IU.ANMO.00.BHZ // Act on channel(s)=BHZ of loc=00 of sta=ANMO of net=IU")
                    logger.info("Example: --level_channel=IU.ANMO.00.*   // Act on channel(s)=*   of loc=00 of sta=ANMO of net=IU")
                    logger.info("Example: --level_channel=IU.ANMO.*.*    // Act on channel(s)=*   of loc=*  of sta=ANMO of net=IU")
                    logger.info("Example: --level_channel=IU.*.*.*       // Act on channel(s)=*   of loc=*  of sta=*    of net=IU")
                    logger.info("Example: --level_channel=*.*.*.BHZ      // Act on channel(s)=BHZ of loc=*  of sta=*    of net=*")
                    parser.print_usage()
                    exit(2)

        (net, sta, loc, cha) = tuple(fields)

        if net == '*': net = None
        if sta == '*': sta = None
        if cha == '*': cha = None
        if loc == '*': loc = None
        if loc == '--': loc = ''

        scnl_filter.NET = net
        scnl_filter.STA = sta
        scnl_filter.CHA = cha
        scnl_filter.LOC = loc

    args.level = level

    if args.epoch_channel is not None:
        scnl_filter.CHN_EPOCH = args.epoch_channel
    if args.epoch_station is not None:
        scnl_filter.STN_EPOCH = args.epoch_station

    return args, scnl_filter

def list_str(values):
    return values.split(',')

class struct:
     def __init__(self, **kwds):
         self.__dict__.update(kwds)

def configure(filename=None):
    """
        Read in config from yaml filename
    """
    configuration = {}
    with open(filename, 'r') as ymlfile:
        configuration = yaml.load(ymlfile, Loader=yaml.FullLoader)
    return configuration



if __name__ == "__main__":
    main()



import os

import logging
logger = logging.getLogger()

from logging.handlers import TimedRotatingFileHandler

def string_to_logLevel(levelString):
    '''
        levelString = 'INFO' or 'WARN' or 'WARNING', etc
    '''
    loglevels = {"ERROR": logging.ERROR, "WARN": logging.WARNING,
                 "WARNING": logging.WARNING, "INFO": logging.INFO,
                 "DEBUG": logging.DEBUG, "CRITICAL": logging.CRITICAL,
                }
    msg = " ".join(list(loglevels.keys()))

    if levelString in loglevels:
        #logger.setLevel(loglevels[levelString])
        return loglevels[levelString]
    else:
        logger.error("Unknown --loglevel=%s (not in:{%s})" % (levelString, msg))
        return None
        # exit(2)

def configure_logger(config, logfile, levelString=None):
    '''
        levelString = 'INFO' or 'WARN' or 'WARNING', etc
    '''
    # DEFAULTS
    DEFAULT_LOG_DIR = os.getcwd()
    DEFAULT_LOG_LEVEL = 'INFO'

    if config:
        LOGDIR = config['LOG_DIR'] if 'LOG_DIR' in config else DEFAULT_LOG_DIR
        log_level = config['LOG_LEVEL'] if 'LOG_LEVEL' in config else DEFAULT_LOG_LEVEL
    else:
        LOGDIR = DEFAULT_LOG_DIR
        log_level = DEFAULT_LOG_LEVEL

    if levelString:
        log_level = levelString

    logger.level = string_to_logLevel(log_level)

    logfile = os.path.join(LOGDIR, logfile)
    if not os.path.exists(LOGDIR):
        try:
            os.makedirs(LOGDIR)
        except:
            raise
    #print("configure_logger: log_file=[%s] level=[%s]" % (logfile, log_level))
    # create file handler 
    fh = TimedRotatingFileHandler(logfile, when="midnight")
    # create console handler 
    ch = logging.StreamHandler()

    #fh.setLevel(logging.DEBUG)
    # create formatter and add it to the handlers
    formatter = logging.Formatter('%(asctime)s [%(levelname)5s] %(message)s')
    ch.setFormatter(formatter)
    fh.setFormatter(formatter)
    # add the handlers to the logger
    logger.addHandler(ch)
    logger.addHandler(fh)

    return



  # ****************************************************************************
  #
  # This file is part of the yasmine editing tool.
  #
  # yasmine (Yet Another Station Metadata INformation Editor), a tool to
  # create and edit station metadata information in FDSN stationXML format,
  # is a common development of IRIS and RESIF.
  # Development and addition of new features is shared and agreed between * IRIS and RESIF.
  #
  #
  # Version 1.0 of the software was funded by SAGE, a major facility fully
  # funded by the National Science Foundation (EAR-1261681-SAGE),
  # development done by ISTI and led by IRIS Data Services.
  # Version 2.0 of the software was funded by CNRS and development led by * RESIF.
  #
  # This program is free software; you can redistribute it
  # and/or modify it under the terms of the GNU Lesser General Public
  # License as published by the Free Software Foundation; either
  # version 3 of the License, or (at your option) any later version. *
  # This program is distributed in the hope that it will be
  # useful, but WITHOUT ANY WARRANTY; without even the implied warranty
  # of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  # GNU Lesser General Public License (GNU-LGPL) for more details. *
  # You should have received a copy of the GNU Lesser General Public
  # License along with this software. If not, see
  # <https://www.gnu.org/licenses/>
  #
  #
  # 2019/10/07 : version 2.0.0 initial commit
  #
  # ****************************************************************************/

import copy
import os
import sys
from sys import exit

import tempfile

from obspy import read_inventory
from obspy.core.inventory.inventory import Inventory
from obspy.core.inventory.channel import Channel
from obspy.core.inventory.network import Network
from obspy.core.inventory.station import Station

import logging
logger = logging.getLogger()

from .. import fdsn_schema_dir, installation_dir

#from libs.libs_xml import valid_xmlfiles, get_schema_version, check_files
from .libs_xml import validate_stationxml, get_schema_version, check_files
from .libs_obs import _write_stationxml
from .plot_poly_resp import plot_polynomial_resp

import threading

def edit_xml_to_inv(args, scnl_filter):
    """
    Read in xml file(s), determine schema version, validate against schema version
        Perform desired modifications (update/delete/add) to metadata
        Return modified metadata in Inventory format

    :param args: Determines what modifications to make to input metadata
    :type args: class argparse.Namespace

    :param scnl_filter: Filter that determines level (network, station, channel) at which to
                        apply modifications
    :type scnl_filter: Simple python struct for holding attributes

    :returns: Modified metadata in obspy inventory format
    :rtype: obspy.core.inventory.inventory

    :returns: Schema_version
    :rtype: string
    """

    fname = 'edit_xml_to_inv'

    cleanup_files = []
    # Verify all input xml file(s) exist
    if args.infiles:
        valid = check_files(args.infiles) # Make sure files exist, are readable, etc.
        if not valid:
            logger.error("One or more xmlfiles could not be read --> STOP EXECUTION")
            exit(2)
    else:
        read_stdin = True
        tf = tempfile.NamedTemporaryFile(delete=False)
        infile = tf.name
        logger.info("Read from stdin infile=[%s] thread=[%s]" % (infile, threading.get_ident()))

        # How to get version direct from stdin either as file or buffer:
        #   If you haven't exhausted sys.stdin, you could do:
        #   schema_version = get_schema_version(sys.stdin)   // read xmlfile to get version
        #   or, you could parse it and do:
        #   buf = ""
        #   for line in sys.stdin:
               #buf += line
        #   schema_version = get_schema_version(buf)         // read from xml_string
        #   schema_version = get_schema_version(infile)      // read from infile

        with open(infile, 'w') as f:
            for line in sys.stdin:
                f.write(line)
        args.infiles = [infile]
        cleanup_files.append(infile)


    # Verify all file(s) have same stationxml schema version:
    versions = []
    for xmlfile in args.infiles:
        versions.append(get_schema_version(xmlfile))
    if len(set(versions)) > 1:
        logger.error("Input files have different schema versions --> Exit")
        exit(2)

    # Validate input xml files against schema
    #          where schema version = --args.schema_version (if set)  *or* schema version of input files
    schema_version = None
    if not args.dont_validate:           # Check for valid StationXML
        valid = True
        for xmlfile in args.infiles:
            schema_version = get_schema_version(xmlfile)
            logger.info("Input schema_version=%s" % schema_version)
            if args.schema_version:
                logger.info("Input files version:[%s] --> Request output version:[%s]" % \
                            (schema_version, args.schema_version))
                schema_version = args.schema_version

            schema_file = os.path.join(fdsn_schema_dir(), 'fdsn-station-%s.xsd' % schema_version)
            logger.info("Check file:%s against schema_file:%s" % (xmlfile, schema_file))
            valid, errors = validate_stationxml(xmlfile, schema_file)

            if not valid:
                for error in errors:
                    logger.error(error)
                logger.error("File:%s does not validate against schema version:[%s]" % \
                            (xmlfile, schema_version))

        if not valid:
            logger.error("One or more xmlfiles are NOT valid StationXML --> STOP EXECUTION")
            exit(2)

    # Read all input xml
    xml_list = load_xmlfiles(args.infiles)

    '''
    for tmpfile in cleanup_files:
        try:
            os.remove(tmpfile)
        except:
            raise
    '''

    if not xml_list:
        logger.error("No xml files loaded --> exit")
        exit(2)

    if args.print_epochs:
        print_all(xml_list, args)
        exit(2)

    # Perform the action
    if args.action == 'delete':
        delete_base_node(xml_list, args.level, scnl_filter)
    elif args.action == 'add':
        add_base_node(xml_list, scnl_filter, args.level, args.value)
    elif args.action == 'update':
        if args.use_index:
            logger.info("Call update_field: field %s[%d] = value=%s" % (args.field, args.field_index, args.value))
        else:
            logger.info("Call update_field: field %s = value=%s" % (args.field, args.value))
        if args.update_root:
            update_root_field(xml_list, args)
        else:
            update_field(xml_list, scnl_filter, args)
    elif args.action == 'select':
        #print("action=select scnl_filter=[%s.%s.%s.%s]" % (scnl_filter.NET, scnl_filter.STA, scnl_filter.LOC, scnl_filter.CHA))
        #filter_xml(xml_list, args.level, scnl_filter)
        filter_xml(xml_list, scnl_filter)
        #exit()

    # Output the modified inventory/stationxml
    inv_new = pack_xml_list_to_inv(xml_list)

    return inv_new, schema_version

#import matplotlib
#matplotlib.use('TkAgg')
#matplotlib.use('agg')
import matplotlib.pyplot as plt

def plot_responses(inventory, plot_dir):
    """
        Loop over channels in inventory, create response plot for each
            channel and save to file in plot_dir
        Response may be 'normal' (= use obspy response.plot()) or
            polynomial (= use plot_polynomial_response() to generate fig)


    """
    if plot_dir is None:
        plot_dir = "."

    else:
        if not os.path.exists(plot_dir):
            try:
                os.makedirs(plot_dir)
            except OSError as e:
                logger.error("Can't create plot_dir: %s --> Check permissions" % (plot_dir))
                exit(2)

    level=logger.level
    logger.setLevel(30)

    min_freq = .001

    logger.info("plot_responses: plot_dir=[%s]" % plot_dir)

    for network in inventory.networks:
        for station in network.stations:
            for channel in station.channels:
                sampling_rate = channel.sample_rate if channel.sample_rate > 0 else 100.

                if channel.start_date is not None:
                    start = channel.start_date.datetime.strftime('%Y-%m-%dT%H:%M:%S')
                    end = channel.end_date.datetime.strftime('%Y-%m-%dT%H:%M:%S') if channel.end_date else None
                    label = "%s.%s.%s.%s.%s-%s" % (network.code, station.code, channel.code,
                                                      channel.location_code, start, end)
                else:
                    label = "%s.%s.%s.%s" % (network.code, station.code, channel.code,
                                                   channel.location_code)
                outfile = label + ".png"
                outfile = os.path.join(plot_dir, outfile)

                if channel.response.instrument_polynomial:
                    #logger.warning("%s.%s.%s.%s has instrument_polynomial: "
                                #"Polynomial response plot not yet supported" %
                                #(network.code, station.code, channel.code, channel.location_code))
                    label += "\npolynomial response"
                    fig = plot_polynomial_resp(channel.response, label=label,
                                               axes=None, outfile=outfile)

                else:
                    fig = channel.response.plot(min_freq, output="VEL", unwrap_phase=False,
                                                sampling_rate=sampling_rate, label=label,
                                                outfile=outfile)
                    plt.close(fig)

    logger.setLevel(level)

    return

def filter_xml(xml_list, scnl_filter):
    """
    Filter (remove) elements defined in scnl_filter from xml_list

    :param xml_list: List of python dicts containing metadata
    :type xml_list: list

    :param scnl_filter: Filter that determines level (network, station, channel) at which to
                        apply modifications
    :type scnl_filter: Simple python struct for holding attributes

    """

    if scnl_filter.NET:
        for xml_dict in xml_list:
            for net_code in xml_dict['net_codes']:
                if net_code != scnl_filter.NET:
                    try:
                        logger.info("Ignore network=%s" % net_code)
                        xml_dict['net_codes'].pop(net_code)
                    except KeyError:
                        logger.error("Key not found:%s" % net_code)
                else:
                    logger.info("Net:%s passed filter" % net_code)

    if scnl_filter.STA:
        for xml_dict in xml_list:
            for net_code, net_dict in xml_dict['net_codes'].items():
                for sta_code in list(net_dict['sta_codes'].keys()):
                    if sta_code != scnl_filter.STA:
                        try:
                            logger.info("Ignore station=%s" % sta_code)
                            net_dict['sta_codes'].pop(sta_code)
                        except KeyError:
                            logger.error("Key not found:%s" % sta_code)
                    else:
                        logger.info("Sta:%s passed filter" % sta_code)

    # At this point you've already filtered net_dict down by NET.STA

    # CHA can be name or *.  LOC can be XX or * or '' (no location)
    if scnl_filter.CHA or scnl_filter.LOC is not None:
        for xml_dict in xml_list:
            #print(xml_dict)
            for net_code, net_dict in xml_dict['net_codes'].items():

                new_dict = {}
                new_dict['sta_codes'] = {}
                new_dict['network'] = net_dict['network']

                for sta_code, sta_epochs in net_dict['sta_codes'].items():
                    stn_epochs = []
                    if not scnl_filter.STA or scnl_filter.STA == sta_code:
                        for station in sta_epochs:
                            channel_epochs = []
                            for ichn, channel in enumerate(station.channels):
                                #print("Here scnl_filter.LOC=[%s]" % scnl_filter.LOC)
                                if (scnl_filter.LOC is None or scnl_filter.LOC == channel.location_code) and \
                                   (scnl_filter.CHA is None or scnl_filter.CHA == channel.code):
                                    logger.info("Sta:%s Cha:%s Loc:%s %s-%s passed filter" %
                                                (sta_code, channel.code, channel.location_code, channel.start_date, channel.end_date))
                                    channel_epochs.append(channel)

                            if channel_epochs:
                                station.channels = channel_epochs
                                stn_epochs.append(station)

                    if stn_epochs:
                        new_dict['sta_codes'][sta_code] = stn_epochs

            xml_dict['net_codes'][net_code] = new_dict


    return


def update_root_field(xml_list, args):
    """
    Update a root field of FDSN StationXML within xml_list

    :param xml_list: List of python dicts containing metadata
    :type xml_list: list

    :param args: Determines what modifications to make to input metadata
    :type args: class argparse.Namespace
    """
    update_pair = args.update_pair
    field = update_pair[0]
    value = update_pair[1]
    for xml in xml_list:
        if field in xml.keys():
            xml[field] = value
            logger.info("Update_root field:%s to value=[%s]" % (field, value))
    return


def update_field(xml_list, scnl_filter, args):
    """
    Update a non-root field of FDSN StationXML within xml_list

    :param xml_list: List of python dicts containing metadata
    :type xml_list: list

    :param scnl_filter: Filter that determines level (network, station, channel) at which to
                        apply modifications
    :type scnl_filter: Simple python struct for holding attributes

    :param args: Determines what modifications to make to input metadata
    :type args: class argparse.Namespace
    """
    update_pair = args.update_pair

    level = args.level

    update_pair = args.update_pair
    field = update_pair[0]
    value = update_pair[1]

    '''
    if basenodetype.attrib is a list and we:
        1. pass in value = list ==> replace old list with new list, even if new list = Empty
            network.operators = value  * This is the same if both attrib and field are scalars!
        2. pass in value = scalar
           if:
               i) --index is not used: append this value to end of old list
                    network.operators.append(value)
              ii) --index is used: replace element at old_list[index] with value
                    network.operators[INDEX] = value
    '''

    def _set_field(basenodeobj, field):

        # MTH: 2021-12-01: obspy v.1.2.2 has a bug in class Identifier @identifiers.setter
        #                  It mistakenly validates identifier value against anyURI which is wrong:
        if field == "identifiers":
            basenodeobj.__setattr__(field, value)
            return True

        if hasattr(basenodeobj, field):
            if args.update in {'list_modify', 'list_append'}:
                current_list = getattr(basenodeobj, field, None)
                if args.update == 'list_append':
                    logger.info("Append to end of list of len:%d" % (len(current_list)))
                    current_list.append(value)
                else: # Modify list (if value=None --> remove item)
                    if len(current_list) <= scnl_filter.INDEX:
                        logger.error("Current_list is of len=%d <= index=%d" % \
                                    (len(current_list), scnl_filter.INDEX))
                        return False
                    if value is None:
                        item = current_list.pop(scnl_filter.INDEX)
                        logger.info("Remove item at INDEX=%d :%s" % (scnl_filter.INDEX, item))

                    else:
                        logger.info("Modify current list index=%d of len:%d" % 
                                   (scnl_filter.INDEX, len(current_list)))
                        current_list[scnl_filter.INDEX] = value
            else:
                logger.info("Simply set field=value. field=%s value=%s" % (field, value))
                try:
                    setattr(basenodeobj, field, value)
                except (TypeError, ValueError) as e:
                    logger.error("Unable to set field:%s to value:%s Caught=[%s]" % \
                                (field, value, repr(e)))
                    return False
        else:
            logger.warning("basenodeobj has no attr=[%s]" % (field))
            return False

        return True

    success = False

    logger.info("update_field: level:%s field:%s value:%s" % (level, field, value))

    if level == 'network':
        for xml_dict in xml_list:
            for net_code, net_dict in xml_dict['net_codes'].items():
                if not scnl_filter.NET or scnl_filter.NET == net_code:
                    network = net_dict['network']
                    logger.info("Update: net:%s ==> field:%s" % (network.code, field))
                    success = _set_field(network, field)

    elif level == 'station':
        for xml_dict in xml_list:
            for net_code, net_dict in xml_dict['net_codes'].items():
                if not scnl_filter.NET or scnl_filter.NET == net_code:
                    for sta_code in list(net_dict['sta_codes'].keys()):
                        if not scnl_filter.STA or scnl_filter.STA == sta_code:
                            for i, station in enumerate(net_dict['sta_codes'][sta_code]):
                                if scnl_filter.STN_EPOCH is None or scnl_filter.STN_EPOCH == i:
                                    logger.info("Update net:%s stn:%s [%d] field:%s" % \
                                                (net_code, sta_code, i, field))
                                    success = _set_field(station, field)
    else:
        for xml_dict in xml_list:
            for net_code, net_dict in xml_dict['net_codes'].items():
                if not scnl_filter.NET or scnl_filter.NET == net_code:
                    for sta_code, station_epochs in net_dict['sta_codes'].items():
                        if not scnl_filter.STA or scnl_filter.STA == sta_code:
                            for istn, station in enumerate(station_epochs):
                                for ichn, channel in enumerate(station.channels):
                                    keycode = "%s.%s" % (channel.code, channel.location_code)
                                    if not scnl_filter.LOC or scnl_filter.LOC == channel.location_code:
                                      if not scnl_filter.CHA or scnl_filter.CHA == channel.code:

                                        if (scnl_filter.STN_EPOCH is None or scnl_filter.STN_EPOCH == istn) and \
                                           (scnl_filter.CHN_EPOCH is None or scnl_filter.CHN_EPOCH == ichn):
                                            logger.info("Update net:%s stn:%s [%d] chn:%s [%d] field:%s" % \
                                                        (net_code, sta_code, istn, keycode, ichn, field))
                                            success = _set_field(channel, field)
    if not success:
        logger.error("Update failed, either because no matching basenodes found or because of error setting attrib")
    return



def add_base_node(xml_list, scnl_filter, level, obj):
    """
    Locate the appropriate position in xml_list and insert a new 
        base node (Network, Station or Channel object).

    :param xml_list: List of python dicts containing metadata
    :type xml_list: list

    :param level: Level at which to apply changes, must be in {'network', 'station', 'channel'}
    :type level: string

    :param scnl_filter: simple python object holding NET, STA, CHA, LOC attributes that mods apply to
    :type scnl_filter: python class used as container

    :param obj: Obspy base_node object (Network, Station, Channel)
    :type obj: obspy.core.inventory.network.Network, obspy.core.inventory.station.Station or obspy.core.inventory.channel.Channel
    """

    if isinstance(obj, Network):
        logger.info("Add Network: Note that level is ignored since Network can only go in root")
        for xml_dict in xml_list:
            if obj.code in xml_dict['net_codes']:
                logger.warning("Warning: Network code:[%s] already exists!" % obj.code)
            else:
                logger.info("Add Network code:[%s] to xml_dict" % obj.code)

                xml_dict['net_codes'][obj.code] = {}
                xml_dict['net_codes'][obj.code]['network'] = obj
                xml_dict['net_codes'][obj.code]['sta_codes'] = {}

# MTH: if obj = Station, then it can go in 1...N networks specified by filter
# Right now the new Station is added to end of each network stations list
# This could be changed to (re)filter the network dicts by alphabet/time before writing xml

    elif isinstance(obj, Station):
        if level != 'network':
            logger.error("Add Station: You *must* use level_network=.. to specify which network(s) get the new station")
            return None
        else:
            for xml_dict in xml_list:
                for net_code, net_dict in xml_dict['net_codes'].items():
                    if not scnl_filter.NET or scnl_filter.NET == net_code:
                    # If this station code already present append a new epoch to its list of Station (epochs)
                        if obj.code in net_dict['sta_codes']:
                            net_dict['sta_codes'][obj.code].append(obj)
                        else:
                            net_dict['sta_codes'][obj.code] = [obj]

# MTH: To add a Channel we need to know which Station(s) will get it
    elif isinstance(obj, Channel):
        if level != 'station':
            logger.error("Add Channel: You *must* use level_station=.. to specify which station(s) get the new channel")
            return None
        else:
            for xml_dict in xml_list:
                for net_code, net_dict in xml_dict['net_codes'].items():
                    if not scnl_filter.NET or scnl_filter.NET == net_code:
                        for sta_code, station_epochs in net_dict['sta_codes'].items():
                            if not scnl_filter.STA or scnl_filter.STA == sta_code:
                                #for station in station_epochs:
                                    #for channel in station.channels:

                                # Inside matching sta_code - which epoch gets it ?
                                station_epochs[-1].channels.append(obj)

    else:
        logger.error("ERROR: Unknown combination: obj type:%s + level" % (type(obj)))

def delete_base_node(xml_list, level, scnl_filter):
    """
    Find the base_node specified via level + scnl_filter and delete from xml_list

    :param xml_list: List of python dicts containing metadata
    :type xml_list: list

    :param level: Level at which to apply changes, must be in {'network', 'station', 'channel'}
    :type level: string

    :param scnl_filter: simple python object holding NET, STA, CHA, LOC attributes that mods apply to
    :type scnl_filter: python class used as container
    """

    if level == 'network':
        for xml_dict in xml_list:
            if scnl_filter.NET in xml_dict['net_codes']:
                try:
                    logger.info("Delete network=%s" % scnl_filter.NET)
                    xml_dict['net_codes'].pop(scnl_filter.NET)
                except KeyError:
                    logger.error("Key not found:%s" % scnl_filter.NET)

    elif level == 'station':
        for xml_dict in xml_list:
            for net_code, net_dict in xml_dict['net_codes'].items():
                if not scnl_filter.NET or scnl_filter.NET == net_code:

                    for sta_code in list(net_dict['sta_codes'].keys()):
                        if not scnl_filter.STA or scnl_filter.STA == sta_code:

                            if scnl_filter.STN_EPOCH is None: # Could have scnl_filter.STN_EPOCH = 0
                                try:
                                    logger.info("Delete net:%s stn:%s all epochs" % (net_code, sta_code))
                                    net_dict['sta_codes'].pop(sta_code)
                                except KeyError:
                                    logger.error("Key not found:%s" % sta_code)
                            else:
                                epochs = net_dict['sta_codes'][sta_code]
                                cleaned_epochs = []
                                for i, epoch in enumerate(epochs):
                                    if i == scnl_filter.STN_EPOCH:
                                        logger.info("Delete net:%s stn:%s epoch:%d" % (net_code, sta_code, i))
                                    else:
                                        cleaned_epochs.append(epoch)

                                net_dict['sta_codes'][sta_code] = cleaned_epochs

    else:
        for xml_dict in xml_list:
            for net_code, net_dict in xml_dict['net_codes'].items():
                if not scnl_filter.NET or scnl_filter.NET == net_code:
                    for sta_code, station_epochs in net_dict['sta_codes'].items():
                        if not scnl_filter.STA or scnl_filter.STA == sta_code:
                            for istn, station in enumerate(station_epochs):
                                cleaned_channels = copy.deepcopy(station.channels)
                                modified = False
                                for ichn, channel in enumerate(station.channels):
                                    if not scnl_filter.LOC or scnl_filter.LOC == channel.location_code:
                                      if not scnl_filter.CHA or scnl_filter.CHA == channel.code:

                                        if (scnl_filter.STN_EPOCH is None or scnl_filter.STN_EPOCH == istn) and \
                                           (scnl_filter.CHN_EPOCH is None or scnl_filter.CHN_EPOCH == ichn):

                                            keycode = "%s.%s" % (channel.code, channel.location_code)
                                            logger.info("Remove net:%s stn:%s [%d] chn:%s [%d] epoch" % \
                                                       (net_code, sta_code, istn, keycode, ichn))
                                            cleaned_channels.remove(channel)
                                            modified = True

                                if modified:
                                    station.channels = cleaned_channels


    return




def pack_xml_list_to_inv(xml_list):
    """
    Convert xml_list of dictionaries containing network(s) metadata to obspy inventory object

    :param xml_list: List of python dicts containing metadata
    :type xml_list: list

    :returns: Inventory object created from xml_list
    :rtype: obspy.core.inventory.inventory
    """

    source = xml_list[0]['source']
    sender = xml_list[0]['sender']
    module = xml_list[0]['module']
    module_uri = xml_list[0]['module_uri']

    inv = Inventory(source=source, sender=sender, module=module, module_uri=module_uri)

    networks = []
    for xml_dict in xml_list:
        for net_code in xml_dict['net_codes']:
            network = xml_dict['net_codes'][net_code]['network']
            stations = []
            for sta_code in xml_dict['net_codes'][net_code]['sta_codes']:
                stations.extend(xml_dict['net_codes'][net_code]['sta_codes'][sta_code])
            network.stations = stations
            networks.append(network)

    inv.networks = networks

    return inv

def load_xmlfiles(xmlfiles):
    """
    Read list of xmlfile(s) into a list of python dicts, one for each xml file,
          where the python dict holds the obspy inventory network objects
          read from the file

    :param xmlfiles: List of xmlfiles with common FDSN schema version
    :type xmlfiles: list

    :returns: list of python dicts
    :rtype: list
    """

    xml_list = []

    for xmlfile in xmlfiles:

        #logger.info("Read_inventory from file:[%s] thread:[%s]" % (xmlfile, threading.get_ident()))

        try:
            inv = read_inventory(xmlfile)
        except ValueError as e:
            logger.error("Problem reading xml file:%s" % repr(e))
            return None

        xml_dict = {}
        xml_list.append(xml_dict)

        xml_dict['xmlfile'] = xmlfile
        xml_dict['source']  = inv.source
        xml_dict['sender']  = inv.sender
        xml_dict['module']  = inv.module
        xml_dict['module_uri'] = inv.module_uri
        xml_dict['net_codes'] = {}

        for network in inv.networks:
            net_dict = network_to_dict(network)
            xml_dict['net_codes'][network.code] = net_dict
            xml_dict['net_codes'][network.code]['network'] = network

    return xml_list


def print_all(xml_list, args):

    for xml_dict in xml_list:
        print("[File:%s]" % xml_dict['xmlfile'])
        for net_code, net_dict in xml_dict['net_codes'].items():
            print("  [Net:%s]" % net_code)
            print_net(net_dict, args)


def print_net(net_dict, args):
    """
    Checks a list of epochs for overlap
        If overlap found, returns 1 + list of overlap messages
        If overlap not found, returns 0 + empty list

    :param epoch_list: List of any like objects with start_date + end_date,
                       Pre-sorted by start_date
    :type epoch_list: list

    :returns: True/False if overlap found
    :rtype: bool

    :returns: List of overlap messages
    :rtype: list
    """

    for stn_code, stn_epochs in net_dict['sta_codes'].items():

        for istn, station in enumerate(stn_epochs):
            #print("stn:%s epoch:%s - %s [nchan=%d]" % \
                  #(station.code, station.start_date, station.end_date, len(station.channels)))
            print("    [Stn:%4s] epoch[%d]:%s - %s" % \
                 (station.code, istn, station.start_date, station.end_date))

            if args.print_all:
                for ic, comment in enumerate(station.comments):
                    print("      [Cmt: %d] %s" % (ic, comment))
                for io, operator in enumerate(station.operators):
                    all_names = []
                    for person in operator.contacts:
                        all_names.extend(person.names)
                    names = " ".join(all_names)
                    all_agency = []
                    for person in operator.contacts:
                        all_agency.extend(person.agencies)
                    agencies = " ".join(all_agency)

                    print("      [ Op: %d] names:%s agencies:%s" % (io, names, agencies))

            for ichan, channel in enumerate(station.channels):
                loc_code = channel.location_code if channel.location_code else '--'
                key_code = "%s.%s" % (channel.code, loc_code)
                #print(key_code)
                print("      [Chn:%s] epoch[%d]:%s - %s" % \
                     (key_code, ichan, channel.start_date, channel.end_date))
                if args.print_all:
                    for ic, comment in enumerate(channel.comments):
                        print("          [Cmt: %d] %s" % (ic, comment))


def network_to_dict(network):
    """
    Convert obspy Network object to compound dict of sorted Station/Channel
        objects (epochs).
        Check for overlapping epochs at both Station + Channel level

    :param network: Network
    :type network: obspy.core.inventory.network.Network

    :returns: Nested dict of sorted epochs
    :rtype: dict
    """

    net_dict = {}
    net_dict['sta_codes'] = {}

    # Sort all station epochs by code + start_date
    sorted_stations = []

    for station in network.stations:
        sorted_stations.append(station)
    sorted_stations.sort(key=lambda x: (x.code, x.start_date), reverse=False)

    # Load up dict by station code
    stn_dict = {}
    for station in sorted_stations:
        if station.code in stn_dict:
            l = stn_dict[station.code]
        else:
            l = []
            stn_dict[station.code] = l
        l.append(station)

    # Store list of station_epochs in net_dict under stn_code
    for stn_code, station_epochs in stn_dict.items():
        net_dict['sta_codes'][stn_code] = station_epochs

        # Check for overlapping station epochs:
        (overlap, msgs) = overlapping_epochs(station_epochs)
        if overlap:
            logger.warning("Stn:%s has overlapping station epochs!" % (station.code))
            for msg in msgs:
                logger.warning(msg)

        # For each station_epoch, sort its channel epochs by keycode + start_date:
        for station in station_epochs:
            sorted_channels = []
            for channel in station.channels:
                sorted_channels.append(channel)
            sorted_channels.sort(key=lambda x: (x.location_code, x.code, x.start_date), reverse=False)

            station.channels = sorted_channels

        # Check for overlapping channel epochs:
            chn_dict = {}
            for channel in sorted_channels:
                keycode = "%s.%s" % (channel.code, channel.location_code)
                if keycode in chn_dict:
                    l = chn_dict[keycode]
                else:
                    l = []
                    chn_dict[keycode] = l
                l.append(channel)

            for keycode, channel_epochs in chn_dict.items():
                (overlap, msgs) = overlapping_epochs(channel_epochs)
                if overlap:
                    logger.error("Stn:%s Chn:%s has overlapping channel epochs!" % (station.code, keycode))
                    for msg in msgs:
                        logger.warning(msg)


    return net_dict


def overlapping_epochs(epoch_list):

    """
    Checks a list of epochs for overlap
        If overlap found, returns 1 + list of overlap messages
        If overlap not found, returns 0 + empty list

    :param epoch_list: List of any like objects with start_date + end_date,
                       Pre-sorted by start_date
    :type epoch_list: list

    :returns: 1 if overlap found else 0
    :rtype: int

    :returns: List of overlap messages
    :rtype: list
    """

    overlap = False

    msgs = []
    #print("overlapping_epochs: list_len=%d" % len(epoch_list))
    for i in range(len(epoch_list)-1):

        epoch1 = epoch_list[i]
        epoch2 = epoch_list[i+1]
        #print("Compare: i=%d" % i)
        #print("  epoch1: %s - %s" % (epoch1.start_date, epoch1.end_date))
        #print("  epoch2: %s - %s" % (epoch2.start_date, epoch2.end_date))

        if epoch1.start_date == epoch2.start_date:
            msgs.append("Epochs have same start_date")
            overlap = True
        # epoch1 better be closed
        elif epoch1.end_date is None:
            msgs.append("Earlier epoch not closed!")
            overlap = True
        # epoch1 close must precede epoch2 start
        elif epoch1.end_date > epoch2.start_date:
            msgs.append("epoch1 end > epoch2 start")
            overlap = True

        if overlap:
            #print("Epoch:%s overlaps with epoch:%s" % (epoch1, epoch2))
            #print("Epoch:%s overlaps with epoch:%s" % (epoch1.start, epoch2))
            msgs.append("These epochs overlap:")
            msgs.append("  epoch1: %s - %s" % (epoch1.start_date, epoch1.end_date))
            msgs.append("  epoch2: %s - %s" % (epoch2.start_date, epoch2.end_date))
            return 1, msgs

    return 0, []



if __name__ == "__main__":
    main()



# ****************************************************************************
  #
  # This file is part of the yasmine editing tool.
  #
  # yasmine (Yet Another Station Metadata INformation Editor), a tool to
  # create and edit station metadata information in FDSN stationXML format,
  # is a common development of IRIS and RESIF.
  # Development and addition of new features is shared and agreed between * IRIS and RESIF.
  #
  #
  # Version 1.0 of the software was funded by SAGE, a major facility fully
  # funded by the National Science Foundation (EAR-1261681-SAGE),
  # development done by ISTI and led by IRIS Data Services.
  # Version 2.0 of the software was funded by CNRS and development led by * RESIF.
  #
  # This program is free software; you can redistribute it
  # and/or modify it under the terms of the GNU Lesser General Public
  # License as published by the Free Software Foundation; either
  # version 3 of the License, or (at your option) any later version. *
  # This program is distributed in the hope that it will be
  # useful, but WITHOUT ANY WARRANTY; without even the implied warranty
  # of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  # GNU Lesser General Public License (GNU-LGPL) for more details. *
  # You should have received a copy of the GNU Lesser General Public
  # License along with this software. If not, see
  # <https://www.gnu.org/licenses/>
  #
  #
  # 2019/10/07 : version 2.0.0 initial commit
  #
  # ****************************************************************************/

import logging
logger = logging.getLogger()

def valid_xmlfiles(args):

    # Verify all input xml a) exist and b) validate
    #schemafile = './fdsn-schema/fdsn-station-1.0.xsd'
    #schemafile = './fdsn-schema/fdsn-station-1.1.xsd'

    if args.infiles:
        valid = validate_files(args.infiles) # Make sure files exist, are readable, etc.
        if not valid:
            logger.error("One or more xmlfiles could not be read --> STOP EXECUTION")
            return False

    if not args.dont_validate:           # Check for valid StationXML
        for xmlfile in args.infiles:
            schema_version = get_schema_version(xmlfile)
            schema_file = 'fdsn-schema/fdsn-station-%s.xsd' % schema_version
            logger.info("Check file:%s against schema_file:%s" % (xmlfile, schema_file))
            #valid, errors = validate_stationxml(BytesIO(contents), schemafile)
            #valid, errors = validate_stationxml(xmlfile, schemafile)
            valid, errors = validate_stationxml(xmlfile, schema_file)

            if not valid:
                for error in errors:
                    logger.error(error)
                logger.error("One or more xmlfiles are NOT valid StationXML --> STOP EXECUTION")
                return False
    return True


from lxml import etree

def validate_stationxml(path_or_object, schemafile):

    try:
      xmlschema = etree.XMLSchema(etree.parse(schemafile))

      if isinstance(path_or_object, etree._Element):
          xmldoc = path_or_object
      else:
          try:
              xmldoc = etree.parse(path_or_object)
          except etree.XMLSyntaxError:
              msg = "file=[%s] is NOT a XML file!" % path_or_object
              return (False, [msg])
    except:
        raise

    valid = xmlschema.validate(xmldoc)

    # Pretty error printing if the validation fails.
    if valid is not True:
        return (False, xmlschema.error_log)

    return (True, ())


import errno
#def validate_files(xmlfiles):
def check_files(xmlfiles):

    for xmlfile in xmlfiles:

        logger.info("Validate exists xml file:[%s]" % xmlfile)

        try:
            with open(xmlfile) as f:
                contents = f.read()
            #f = xmlfile
            #contents = f.read()
        except IOError as x:
            if x.errno == errno.ENOENT:
                logger.error('File:%s does not exist' % xmlfile)
            elif x.errno == errno.EACCES:
                logger.error('File:%s Not readable' % xmlfile)
            else:
                logger.error('File:%s Caught error:%s' % (xmlfile, repr(x)))
            return False

    return True


import xml.etree.ElementTree as ET
#from lxml import etree as ET
def get_schema_version(xmlfile_or_string):

    namespace = "http://www.fdsn.org/xml/station/1"

    root = None
    try:
        # parse an xmlfile
        tree = ET.parse(xmlfile_or_string)
        root = tree.getroot()
    except:
        # parse an xml string
        root = ET.fromstring(xmlfile_or_string)
    if root is not None:
        stationxml_version = root.attrib.get('schemaVersion')
        return stationxml_version
    else:
        logger.ERROR("Unable to parse xmlfile_or_string")
        return None


def set_schema_version(xmlfile, version):

    namespace = "http://www.fdsn.org/xml/station/1"

    try:
        tree = ET.parse(xmlfile)
        root = tree.getroot()
        root.attrib['schemaVersion'] = version
    except:
        raise

    return tree





#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_yasmine_cli
----------------------------------

Tests for `yasmine_cli` module.
"""

import glob
import os
import sys
import unittest

from obspy.core.inventory.network import Network
from obspy.core.inventory.station import Station
from obspy.core.inventory.channel import Channel
from obspy import read_inventory

import yasmine_cli

from yasmine_cli import installation_dir, fdsn_schema_dir, yml_template_dir
from yasmine_cli.libs.libs_xml import validate_stationxml, get_schema_version
from yasmine_cli.libs.edit_xml_to_inv import load_xmlfiles, pack_xml_list_to_inv
from yasmine_cli.libs.libs_obs import _write_stationxml, read_yml_file
from yasmine_cli.libs.libs_util import configure, processCmdLine
from yasmine_cli.libs.libs_log import configure_logger
from yasmine_cli.libs.edit_xml_to_inv import update_root_field, update_field, edit_xml_to_inv

import logging
logger = logging.getLogger()
logger.setLevel(logging.ERROR)

TEMPLATE_DIR = yml_template_dir()

class TestYasmine_cli(unittest.TestCase):

    def setUp(self):

        self.install_dir = installation_dir()
        self.assertEqual(os.path.isdir(self.install_dir), 1)
        self.schema_dir  = fdsn_schema_dir()
        self.assertEqual(os.path.isdir(self.schema_dir), 1)

    def test_version(self):
        assert(yasmine_cli.__version__)

    def test_read_config(self):
        config = configure('./config.yml')
        self.assertEqual(config['LOG_LEVEL'], 'DEBUG')

    def test_process_command_line(self):
        #sys.argv = ['/Users/mth/mth/miniconda3/envs/yas/bin/yasmine-cli',
        sys.argv = ['yasmine-cli',
                    '--field=code', '--value=MIKE',
                    '--level_station=*.ANMO', '--epoch_station=1', '--epoch_channel=3',
                    '--schema_version', '1.1',
                    '-o', 'a.xml']
        args, scnl_filter = processCmdLine('yasmine-cli')
        self.assertEqual(args.level, 'station')
        self.assertEqual(args.field, 'code')
        self.assertEqual(args.value, 'MIKE')
        self.assertEqual(scnl_filter.STA, 'ANMO')
        self.assertEqual(scnl_filter.STN_EPOCH, 1)
        self.assertEqual(scnl_filter.CHN_EPOCH, 3)
        pass

    def test_read_xml_version(self):
        xmlfile = 'test_data/Test.xml'
        schema_version = get_schema_version(xmlfile)
        self.assertEqual(schema_version, '1.0')
        schema_file = os.path.join(self.schema_dir, 'fdsn-station-%s.xsd' % schema_version)
        assert(os.path.isfile(schema_file))
        valid, errors = validate_stationxml(xmlfile, schema_file)
        self.assertTrue(valid)

    def test_load_xml_file(self):
        xmlfile = 'test_data/Test.xml'
        xml_list = load_xmlfiles([xmlfile])
        self.assertIsInstance(xml_list, list)

    def test_convert_xml_list_to_inv(self):
        xmlfile = 'test_data/Test.xml'
        outfile = 'b.xml'
        xml_list = load_xmlfiles([xmlfile])
        inv = pack_xml_list_to_inv(xml_list)
        schema_version = get_schema_version(xmlfile)
        self.assertIsNotNone(inv)
        _write_stationxml(inv, outfile, validate=True, schema_version=schema_version)
        assert(os.path.isfile(outfile))

    def test_read_base_nodes(self):
        for basenode in [Network, Station, Channel]:
            node = test_read_base_node(basenode.__name__)
            self.assertIsInstance(node, basenode)

    def test_update_source(self):
        sys.argv = ['yasmine-cli', '--infiles=test_data/station.xml', '-o', 'a.xml',
                    '--field=source', '--value=RESIF-RAP',
                    ]
        args, scnl_filter = processCmdLine('yasmine-cli')
        inv, schema_version = edit_xml_to_inv(args, scnl_filter)
        _write_stationxml(inv, args.output, validate=True, schema_version=schema_version)

    def test_add_channel(self):
        sys.argv = ['yasmine-cli', '--infiles=test_data/Test.xml', '-o', '1.xml',
                    '--action=add', '--from_yml=yml:yml/channel.yml',
                    '--level_station=*.ANMO',
                    ]
        args, scnl_filter = processCmdLine('yasmine-cli')
        inv, schema_version = edit_xml_to_inv(args, scnl_filter)
        _write_stationxml(inv, args.output, validate=True, schema_version=schema_version)

    def test_add_station(self):
        sys.argv = ['yasmine-cli', '--infiles=test_data/NE.xml', '-o', '2.xml',
                    '--action=add', '--from_yml=yml:yml/station.yml',
                    '--level_network=NE',
                    '--schema_version=1.1',
                    ]
        args, scnl_filter = processCmdLine('yasmine-cli')
        inv, schema_version = edit_xml_to_inv(args, scnl_filter)
        _write_stationxml(inv, args.output, validate=True, schema_version=schema_version)

    def test_add_site(self):
        sys.argv = ['yasmine-cli', '--infiles=test_data/Test.xml', '-o', '3.xml',
                    #'--action=update',
                    '--field=site', '--value=yml:yml/site.yml',
                    '--level_station=*.ANMO',
                    '--schema_version=1.1',
                    ]
        args, scnl_filter = processCmdLine('yasmine-cli')
        inv, schema_version = edit_xml_to_inv(args, scnl_filter)
        _write_stationxml(inv, args.output, validate=True, schema_version=schema_version)
        inv = read_inventory('3.xml')
        tmp = inv.select(station="ANMO")
        self.assertEqual(tmp.networks[0].stations[0].site.town, 'Isleta Pueblo')
        self.assertEqual(tmp.networks[0].stations[0].site.county, 'Bernalillo')

    def test_pipe_0(self):
        """Piping example:
            1) Change all ANMO station codes to 'MIKE',
            2) Change the latitude of all CCM stations to 33.77,
            3) Add list of operators to all MIKE stations, and
            4) Replace the 2nd operator in the list with a new operator from yml file:
        """

        cmd = ("cat test_data/Test.xml | yasmine-cli --field=code --value=MIKE --level_station=*.ANMO --dont_validate "
               "| yasmine-cli --field=latitude --value=33.77 --level_station=*.CCM "
               "| yasmine-cli --field=operators --value=yml:yml/operators.yml --level_station=*.MIKE "
               "| yasmine-cli --field=operators[1] --value=yml:yml/operator.yml --level_station=*.MIKE -o y.xml "
               )
        os.system(cmd)

    def test_pipe_1(self):

        inv = read_inventory('y.xml')
        tmp = inv.select(station="CCM")
        self.assertEqual(tmp.networks[0].stations[0].latitude, 33.77)

        tmp = inv.select(station="MIKE")
        for station in tmp.networks[0].stations:
            self.assertEqual(station.operators[0].agency, 'United States Geological Survey, USGS')
            self.assertEqual(station.operators[0].contacts[0].names[0], 'Adam Ringler')
            self.assertEqual(station.operators[0].contacts[0].agencies[0], 'Albuquerque Seismic Lab')
            self.assertEqual(station.operators[1].agency, 'UCSC')
            self.assertEqual(station.operators[1].contacts[0].names[0], 'Dan Sampson')
            self.assertEqual(station.operators[1].contacts[0].emails[0], 'd.sampson@ucsc.edu')


    def test_update_0(self):
        sys.argv = ['yasmine-cli', '--infiles=test_data/station.xml', '-o', '4.xml',
                    '--action=update', '--level_station=*.*',
                    '--field=comments', '--value=yml:yml/comment.yml',
                    ]
        args, scnl_filter = processCmdLine('yasmine-cli')
        inv, schema_version = edit_xml_to_inv(args, scnl_filter)
        _write_stationxml(inv, args.output, validate=True, schema_version=schema_version)
        inv = read_inventory('4.xml')
        #print(inv.networks[0].stations[0].comments[0].authors[0].names[0])
        #print(inv.networks[0].stations[0].comments[0].authors[0].agencies[0])
        self.assertEqual(inv.networks[0].stations[0].comments[0].authors[0].names[0], 'Jean-Marie Saurel')
        self.assertEqual(inv.networks[0].stations[0].comments[0].authors[0].agencies[0], 'Institut de Physique du Globe de Paris')


    def test_update_1(self):
        """Replace the 7th comment of the 128th channel epoch of the
            5th station epoch with a comment from yaml file
        """
        sys.argv = ['yasmine-cli', '--infiles=test_data/ANMO.xml', '-o', '5.xml',
                    '--action=update', '--level_channel=*.ANMO.10.VHZ',
                    '--epoch_channel=128', '--epoch_station=5',
                    '--field=comments[6]', '--value=yml:yml/comment.yml',
                    '--dont_validate'
                    ]

        args, scnl_filter = processCmdLine('yasmine-cli')
        inv, schema_version = edit_xml_to_inv(args, scnl_filter)
        _write_stationxml(inv, args.output, validate=False, schema_version=schema_version)
        inv = read_inventory('5.xml')

        self.assertEqual(inv.networks[0].stations[5].channels[128].comments[6].authors[0].names[0],
                         'Jean-Marie Saurel')
        self.assertEqual(inv.networks[0].stations[5].channels[128].comments[6].authors[1].names[0],
                         'Sid Hellman')

    def test_update_3(self):
        """Remove the 0th comment from the 128th channel epoch of the
            5th station epoch, then read it back in and make sure
            what was the comment at channel.coments[6] is now at channel.comments[5]
        """
        #sys.argv = ['yasmine-cli', '--infiles=test_data/ANMO.xml', '-o', '6.xml',
        sys.argv = ['yasmine-cli', '--infiles=5.xml', '-o', '6.xml',
                    '--action=update', '--level_channel=*.ANMO.10.VHZ',
                    '--epoch_channel=128', '--epoch_station=5',
                    '--field=comments[0]', '--value=None',
                    '--dont_validate'
                    ]

        args, scnl_filter = processCmdLine('yasmine-cli')
        inv, schema_version = edit_xml_to_inv(args, scnl_filter)
        _write_stationxml(inv, args.output, validate=False, schema_version=schema_version)
        inv = read_inventory('6.xml')
        self.assertEqual(len(inv.networks[0].stations[5].channels[128].comments), 6)
        #for comment in inv.networks[0].stations[5].channels[128].comments:
            #print(comment.value)
        self.assertEqual(inv.networks[0].stations[5].channels[128].comments[5].authors[0].names[0],
                         'Jean-Marie Saurel')
        self.assertEqual(inv.networks[0].stations[5].channels[128].comments[5].authors[1].names[0],
                         'Sid Hellman')


    @classmethod
    def tearDownClass(cls):
        print("Call teardownclass")
        files = glob.glob('?.xml')
        for f in files:
            os.remove(f)


def test_read_base_node(basenode):
        yml_path = 'yml:yml/%s.yml' % basenode.lower()
        file_1 = os.path.join(os.getcwd(), yml_path[4:])
        # Else look for file in TEMPLATE_DIR
        file_2 = os.path.join(TEMPLATE_DIR, os.path.basename(yml_path[4:]))

        if os.path.exists(file_1):
            ymlfile = file_1
        elif os.path.exists(file_2):
            ymlfile = file_2
        else:
            print("Unable to find yml_file=[%s] file in either: cwd=%s -or: TEMPLATE_DIR=%s" %
                  (yml_path[4:], os.getcwd(), TEMPLATE_DIR))

        return read_yml_file(ymlfile)


if __name__ == "__main__":
    unittest.main()




from obspy.core.inventory.response import PolynomialResponseStage, InstrumentPolynomial
import matplotlib.pyplot as plt
import os
import numpy as np

import logging
logger = logging.getLogger()

def plot_polynomial_resp(response, label=None, axes=None, outfile=None,
                         vmin=-20., vmax=20., dv=0.10):
    """
    Plot polynomial response
        The way that polynomial responses are calculated is a bit "bassackwards":
        Given a range of inputs (e.g., temperatures), we use the polynomial coefficients
        to calculate what the corresponding outputs (volts or counts) would be.
        We plot input_units on the x-axis and output_units on the y-axis to form y=y(x).

        Will produce 2 subplots: 1. xy-plot of y=output_units (e.g., Volts) vs x=input_units (e.g., degC)
                                 2. xy-plot of y=output_units (e.g., Counts)vs x=input_units (e.g., degC)

        For a MCLAUREN polynomial response, plot 1 comes from the PolynomialResponseStage (e.g., first stage)
              while plot 2 comes from the InstrumentPolynomial (overall response with net gain included in coeffs)

    :param response: channel polynomial response
    :type output: ObsPy response object

    MTH: By default, plot_polynomial_resp will step over voltage from -20V to +20V
         with dV step=0.1V, but the final plot xy ranges will be set by
         the input (x-axis) range: poly.approximation_lower_bound - upper_bound
         However, I left vmin/vmax/dv configurable in case a calling function wants
         to control this to limit plot range (somehow).
    """

    if not response.instrument_polynomial or not response.response_stages or not isinstance(response.response_stages[0], PolynomialResponseStage):
        logger.error("plot_polynomial_resp: response does not contain instrument_polynomial "
                     "and/or PolynomialResponseStage")
        return None

    #file_name = outfile
    #os.makedirs(folder, exist_ok=True)
    #sanitized_file_name = file_name.replace('/', '_').replace('\\', '_') + '.png'
    #file_path = os.path.join(folder, f'{sanitized_file_name}')
    #outfile = file_path

    # We'll use the overall gain to scale between Volts and Counts
    net_gain = 1.
    for i, stage in enumerate(response.response_stages):
        if stage.stage_gain:
            net_gain *= stage.stage_gain

    poly = response.response_stages[0]
    xlabel = poly.input_units
    ylabel = poly.output_units

    # MTH: We need a min/max in the *output* space (e.g., volts) to step through
    #      Otherwise we'll be calculating wild values out of range
    #      Add a little bit to vmax to it gets included in the array
    volts = np.arange(vmin, vmax+dv/10., dv)

    # Load x=temp = f(y=volts) // the measured thing might not be "temp", it doesn't matter
    x1 = []
    y1 = []
    for volt in volts:
        temp = 0.
        for i, c in enumerate(poly.coefficients):
            temp += c * np.power(volt, i)
        if temp >= poly.approximation_lower_bound and temp <= poly.approximation_upper_bound:
            # print("V:%.3f     T:%.2f" % (volt, temp))
            x1.append(temp)
            y1.append(volt)

    # Load x=temp = f(y=counts)
    poly = response.instrument_polynomial
    y2label = poly.output_units
    x2 = []
    y2 = []
    print()
    for volt in volts:
        temp = 0.
        count = volt * net_gain
        for i, c in enumerate(poly.coefficients):
            temp += c * np.power(count, i)
        if temp >= poly.approximation_lower_bound and temp <= poly.approximation_upper_bound:
            # print("C:%.3f     T:%.2f" % (count, temp))
            x2.append(temp)
            y2.append(count)

    if axes:
        ax1, ax2 = axes
        fig = ax1.figure
    else:
        fig = plt.figure()
        ax1 = fig.add_subplot(211)
        ax2 = fig.add_subplot(212, sharex=ax1)

    label_kwarg = {}
    #if label is not None:
       #label_kwarg['label'] = label

    plt.suptitle(label, size='small')
    plt.xlabel(xlabel)
    ax1.set_ylabel(ylabel)
    ax2.set_ylabel(y2label)
    ax2.ticklabel_format(axis='y', style='sci', useMathText=True, scilimits=(0, 0))

    lw = 1.5
    marker = "."
    color = 'red'
    lines = ax1.plot(x1, y1, marker=marker, color=color, **label_kwarg)

    # color = lines[0].get_color()
    lines = ax2.plot(x2, y2, color=color, marker=marker)

    plt.subplots_adjust(left=None, bottom=None, right=None, top=None, wspace=None, hspace=0.3)

    show = 1
    if outfile:
        fig.savefig(outfile)
    else:
        if show:
            plt.show()

    return fig
    #return sanitized_file_name





"""A setuptools based setup module for yasmine-cli"""
#!/usr/bin/env python
# -*- coding: utf-8 -*-

from codecs import open
import os
from os import path
from setuptools import setup, find_packages

#import versioneer
import yasmine_cli

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as readme_file:
    readme = readme_file.read()

with open(path.join(here, 'HISTORY.rst'), encoding='utf-8') as history_file:
    history = history_file.read().replace('.. :changelog:', '')

requirements = [
    'obspy>=1.2',
    'PyYAML',
]

test_requirements = [
    # TODO: put package test requirements here
    'tox',
    'pytest',
    'pytest-runner',
]

import sys
import site
#print("***** MTH: sys.prefix=[%s] site.USER_BASE=[%s]" % (sys.prefix, site.USER_BASE))
py_version = '%s.%s' % (sys.version_info[0], sys.version_info[1])
path = 'lib/python%s/site-packages/yasmine_cli' % py_version
data_dir = os.path.join(path, 'fdsn-schema')
data_files = [(data_dir, ["yasmine_cli/fdsn-schema/fdsn-station-1.0.xsd",
                          "yasmine_cli/fdsn-schema/fdsn-station-1.1.xsd"]
             )]

setup(
    name='yasmine-cli',
    #version=versioneer.get_version(),
    #cmdclass=versioneer.get_cmdclass(),
    version=yasmine_cli.__version__,
    description="Command line tool for editing StationXML files",
    #long_description=readme + '\n\n' + history,
    long_description=open('README.rst', 'r').read(),
    long_description_content_type='text/x-rst',
    author="Mike Hagerty",
    author_email='mhagerty@isti.com',
    url='https://gitlab.isti.com/mhagerty/yasmine-cli',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    #data_files = [("yasmine_cli", data_files)],
    data_files = data_files,
    entry_points={
        'console_scripts':[
            'yasmine-cli=yasmine_cli.yasmine_cli:main',
            ],
        },
    include_package_data=True,
    package_data = { 'yasmine_cli': ['yml/*', 'config.yml'] },
    install_requires=requirements,
    license="MIT",
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements,
)



# ****************************************************************************
  #
  # This file is part of the yasmine editing tool.
  #
  # yasmine (Yet Another Station Metadata INformation Editor), a tool to
  # create and edit station metadata information in FDSN stationXML format,
  # is a common development of IRIS and RESIF.
  # Development and addition of new features is shared and agreed between * IRIS and RESIF.
  #
  #
  # Version 1.0 of the software was funded by SAGE, a major facility fully
  # funded by the National Science Foundation (EAR-1261681-SAGE),
  # development done by ISTI and led by IRIS Data Services.
  # Version 2.0 of the software was funded by CNRS and development led by * RESIF.
  #
  # This program is free software; you can redistribute it
  # and/or modify it under the terms of the GNU Lesser General Public
  # License as published by the Free Software Foundation; either
  # version 3 of the License, or (at your option) any later version. *
  # This program is distributed in the hope that it will be
  # useful, but WITHOUT ANY WARRANTY; without even the implied warranty
  # of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  # GNU Lesser General Public License (GNU-LGPL) for more details. *
  # You should have received a copy of the GNU Lesser General Public
  # License along with this software. If not, see
  # <https://www.gnu.org/licenses/>
  #
  #
  # 2019/10/07 : version 2.0.0 initial commit
  #
  # ****************************************************************************/

import copy
import os
import yaml

from obspy.core.inventory.util import (DataAvailability, DataAvailabilitySpan,
                                       Equipment, Operator, Person,
                                       PhoneNumber, ExternalReference, Comment,
                                       Site,
                                       )

from obspy.core.inventory.channel import Channel
from obspy.core.inventory.network import Network
from obspy.core.inventory.station import Station

import importlib
utils_module = importlib.import_module("obspy.core.inventory.util")

import logging
logger = logging.getLogger()

def main():
    show_fields()

    return


inventory_fields = {'source':'IRIS-DMC', 'sender':'IRIS-DMC',
                    'module':'IRIS WEB SERVICE: fdsnws-station | version: 1.1.41',
                    'module_uri' : 'https://seiscode.iris.washington.edu/projects/stationxml-converter/wiki',
                    }

network_fields = {'code':'IU', 'alternate_code':'XX', 'historical_code':'YY',
                  'description':"'This is a network'",
                  'restricted_status':'open',
                  'start_date':'2010-09-01T00:00:00', 'end_date':'2015-03-01T00:00:00',
                  'selected_number_of_stations':162,
                  'total_number_of_stations':195,
                  'source_id':"'example schema:path'",
                  'identifiers':"['id_1 schema:path', 'id_2 schema:path', ...]",
                  'operators':'yml:yml/operators.yml', 'comments':'yml:yml/comments.yml',
                  'data_availability':'yml:yml/data_availability.yml',
                 }

station_fields = {'code':'ANMO', 'alternate_code':'AAAA', 'historical_code':'BBBB',
                  'description':"'this is a station'",
                  'restricted_status':'open',
                  'start_date':'2010-09-01T00:00:00', 'end_date':'2015-03-01T00:00:00',
                  'creation_date':'2010-09-01T00:00:00', 'termination_date':'2010-09-01T00:00:00',
                  'latitude':34.9459, 'longitude':-106.4572, 'elevation':1850.0,
                  'selected_number_of_channels':27,'total_number_of_channels':42,
                  'vault':"'This is a vault string'", 'geology':"'This is a geology string'",
                  'source_id':"'example schema:path'",
                  'identifiers':"['id_1 schema:path', 'id_2 schema:path', ...]",
                  'operators':'yml:yml/operators.yml', 'comments':'yml:yml/comments.yml',
                  'data_availability':'yml:yml/data_availability.yml',
                  'equipments':'yml:yml/equipments.yml',
                  'site':'yml:yml/site.yml', 'external_references':'yml:yml/references.yml'
                 }

channel_fields = {'code':'BHZ', 'location_code':'00', 'location_code':'XXX', 'historical_code':'YYY',
                  'description':"'this is a channel'",
                  'restricted_status':'open',
                  'start_date':'2010-09-01T00:00:00', 'end_date':'2015-03-01T00:00:00',
                  'latitude':34.9459, 'longitude':-106.4572, 'elevation':1850.0, 'depth':100.0,
                  'azimuth':0.0, 'dip':90.0,
                  'water_level':120.0,
                  'types':"['CONTINUOUS', 'GEOPHYSICAL', ...]",

                  'sample_rate':40.0,
                  'sample_rate_ratio_number_samples':40,
                  'sample_rate_ratio_number_seconds':40,
                  'clock_drift_in_seconds_per_sample':0.01,
                  'calibration_units':"'M/S'",
                  'calibration_units_description':"'Velocity in meters per second'",

                  'source_id':"'example schema:path'",
                  'identifiers':"['id_1 schema:path', 'id_2 schema:path', ...]",
                  'operators':'yml:yml/operators.yml', 'comments':'yml:yml/comments.yml',
                  'data_availability':'yml:yml/data_availability.yml',

                  'sensor':'yml:yml/sensor.yml',
                  'data_logger':'yml:yml/data_logger.yml',
                  'pre_amplifier':'yml:yml/pre_amplifier.yml',
                  'equipments':'yml:yml/equipments.yml',

                  'external_references':'yml:yml/references.yml'
                 }
def show_fields():
    """
    yasmine-cli --show_fields  or
    yasmine-cli --show-fields

    For each level (Network, Station, etc), prints out the fields that can be 
        changed with this code
    """

    print("               root: [=default level] ")
    print("%40s    %s" % ('--field', '--value'))
    print("%40s    %s" % ('==========', '=========='))
    for k,v in inventory_fields.items():
        print("%40s    %s" % (k,v))


    print("    --level_network:")
    print("%40s    %s" % ('--field', '--value'))
    print("%40s    %s" % ('==========', '=========='))
    for k,v in network_fields.items():
        print("%40s    %s" % (k,v))

    print()
    print("    --level_station:")
    print("%40s    %s" % ('--field', '--value'))
    print("%40s    %s" % ('==========', '=========='))
    for k,v in station_fields.items():
        print("%40s    %s" % (k,v))

    print()
    print("    --level_channel:")
    print("%40s    %s" % ('--field', '--value'))
    print("%40s    %s" % ('==========', '=========='))
    for k,v in channel_fields.items():
        print("%40s    %s" % (k,v))

    return

def check_field(field, value, level):
    """
    Confirm that field is in obj_dict at level

    :param field:
    :type field:

    :param value:
    :type value:

    :param level: Level at which to apply changes, must be in {'network', 'station', 'channel'}
    :type level: string

    :returns: True if field in obj_dict at specified level
    :rtype: bool
    """

    obj_dict = "%s_fields" % level
    if field not in eval(obj_dict):
        return False
    return True

def read_yml_file(configFile):
    """
    Read in configFile and use it to create some class
        of obspy object (e.g., Network/Station/Channel/Operator/Comment)

    Return the newly created obspy object

    :param configFile: Name of yml file holding template for particular obspy object
    :type configFile: str

    :returns: obspy object made from template defined in configFile yml
    :rtype: obspy object
    """

    # ObsPy objects that contain another ObsPy object
    #obspy_complex_types = ['Operator', 'Comment', 'DataAvailability']
    obspy_complex_types = ['Operator', 'Comment', 'DataAvailability', 'Identifier']

    return_obj = None

    with open(configFile, 'r') as ymlfile:
        cfg = yaml.load(ymlfile, Loader=yaml.FullLoader)

    first_key = list(cfg.keys())[0]

    process_list = False
    if isinstance(cfg[first_key], list):
        obj_key = list(cfg[first_key][0].keys())[0]
        process_list = True
    else:
        obj_key = first_key

    if not process_list:
        obj_dict = cfg[first_key]

        if obj_key in obspy_complex_types:
            func = 'obspy_%s' % obj_key.lower()
        else:
            func = 'obspy_simple_type'

        return eval(func)(obj_key, obj_dict)

    else:
        #for obj_dict in cfg[first_key][obj_key]:
        return_list = []
        for item in cfg[first_key]:
            obj_dict = item[obj_key]
            if obj_key in obspy_complex_types:
                func = 'obspy_%s' % obj_key.lower()
            else:
                func = 'obspy_simple_type'
            return_list.append(eval(func)(obj_key, obj_dict))

        #field = obj_key.lower() + 's'
        field = obj_key.lower()
        #return return_list, field
        return return_list


def obspy_identifier(obj_key, obj_dict):
    # MTH: there is no Identifier class in obspy
    #      so here we simple return string of value and append to list
    return str(obj_dict)

def obspy_comment(obj_key, obj_dict):
    """
    Create obspy Comment object from fields in obj_dict

    :param obj_dict: fields needed to create an obspy Comment
    :type obj_dict: dict

    :returns: obspy Comment object
    :rtype: obspy.core.inventory.util.Comment
    """

    persons =[]
    for contact in obj_dict['authors']:
        persons.append(obspy_person('Person', contact['author']))

    obj_dict['authors'] = persons
    #comment_text = obj_dict.pop('text')
    #obj_dict['value'] = comment_text
    obj_dict['value'] = obj_dict.pop('value')

    return _make_obj(obj_key, obj_dict)


def obspy_dataavailability(obj_key, obj_dict):
    """
    Create obspy DataAvailability object from fields in obj_dict

    :param obj_dict: fields needed to create an obspy DataAvailability
    :type obj_dict: dict

    :returns: obspy DataAvailability object
    :rtype: obspy.core.inventory.util.DataAvailability
    """

    if 'spans' in obj_dict and obj_dict['spans']:
        spans = []
        for span in obj_dict['spans']:
            spans.append(DataAvailabilitySpan(**span))
        obj_dict['spans'] = spans

    return _make_obj(obj_key, obj_dict)


def obspy_operator(obj_key, obj_dict):
    """
    Create obspy Operator object from fields in obj_dict

    :param obj_dict: fields needed to create an obspy Operator
    :type obj_dict: dict

    :returns: obspy Operator object
    :rtype: obspy.core.inventory.util.Operator
    """

    if 'agency' not in obj_dict or obj_dict['agency'] is None:
        logger.error("Operator agency field is required! Please set in yml file")
        exit(2)

    if 'contacts' in obj_dict:
        persons =[]
        for contact in obj_dict['contacts']:
            persons.append(obspy_person('Person', contact['person']))
        obj_dict['contacts'] = persons

    return _make_obj(obj_key, obj_dict)


def obspy_person(obj_key, obj_dict):
    """
    Create obspy Person object from fields in obj_dict

    :param obj_dict: fields needed to create an obspy Person
    :type obj_dict: dict

    :returns: obspy Person object
    :rtype: obspy.core.inventory.util.Person
    """

    phonenumbers = []
    if 'phones' in obj_dict and obj_dict['phones']:
        for phone in obj_dict['phones']:
            phonenumbers.append(PhoneNumber(phone[1:4], phone[5:], description=None))
        obj_dict['phones'] = phonenumbers

    return _make_obj(obj_key, obj_dict)


def obspy_simple_type(obj_key, obj_dict):
    """
    Create instance of obspy object with class name = obj_key
        (e.g., Network/Station/Channel/Operator/Comment/Equipment/etc)
        using args contained in obj_dict

    :param obj_dict: python dict with args for created obspy object
    :type obj_dict: dict

    :returns: obspy object of class obj_key
    :rtype: obspy object
    """

    return _make_obj(obj_key, obj_dict)


def _make_obj(obj_key, obj_dict):
    """
    Create instance of obspy object with class name = obj_key
    using params in obj_dict

    :param obj_key: str (=class name) of obspy class
    :type obj_key: str

    :param obj_dict: python dict with args for created obspy object
    :type obj_dict: dict

    :returns: obspy object of class obj_key
    :rtype: obspy object
    """

    try:
        #obj = getattr(utils_module, obj_key)(**obj_dict)
        obj = eval(obj_key)(**obj_dict)
    except TypeError as e:
        logger.error("Caught: %s" % repr(e))
        raise
    #return obj, obj_key.lower()
    return obj


# MTH: Everything below here is a quick hack to override the hard-coded
#      schema version (SCHEMA_VERSION = '1.1') in ObsPy inventory module.
#      The only change is looking through kwargs for 'schema_version='

from lxml import etree

from obspy.io.stationxml.core import _write_network, _write_extra
from obspy.io.stationxml.core import validate_stationxml

import io

def _write_stationxml(inventory, file_or_file_object, validate=False,
                      nsmap=None, level="response", **kwargs):
    """
    Writes an inventory object to a buffer.
    :type inventory: :class:`~obspy.core.inventory.Inventory`
    :param inventory: The inventory instance to be written.
    :param file_or_file_object: The file or file-like object to be written to.
    :type validate: bool
    :param validate: If True, the created document will be validated with the
        StationXML schema before being written. Useful for debugging or if you
        don't trust ObsPy. Defaults to False.
    :type nsmap: dict
    :param nsmap: Additional custom namespace abbreviation mappings
        (e.g. `{"edb": "http://erdbeben-in-bayern.de/xmlns/0.1"}`).

    """

    SCHEMA_VERSION = '1.1'

    if 'schema_version' in kwargs and kwargs['schema_version'] is not None:
        SCHEMA_VERSION = kwargs['schema_version']
        logger.info("_write_stationxml: set output schema_version=[%s]" % SCHEMA_VERSION)

    if nsmap is None:
        nsmap = {}
    elif None in nsmap:
        msg = ("Custom namespace mappings do not allow redefinition of "
               "default StationXML namespace (key `None`). "
               "Use other namespace abbreviations for custom namespace tags.")
        raise ValueError(msg)

    nsmap[None] = "http://www.fdsn.org/xml/station/1"
    attrib = {"schemaVersion": SCHEMA_VERSION}

    root = etree.Element("FDSNStationXML", attrib=attrib, nsmap=nsmap)

    etree.SubElement(root, "Source").text = inventory.source
    if inventory.sender:
        etree.SubElement(root, "Sender").text = inventory.sender

    # Undocumented flag that does not write the module flags. Useful for
    # testing. It is undocumented because it should not be used publicly.
    if kwargs.get("_suppress_module_tags", False):
        pass
    else:
        etree.SubElement(root, "Module").text = inventory.module
        etree.SubElement(root, "ModuleURI").text = inventory.module_uri

    etree.SubElement(root, "Created").text = str(inventory.created)

    if level not in ["network", "station", "channel", "response"]:
        raise ValueError("Requested stationXML write level is unsupported.")

    for network in inventory.networks:
        _write_network(root, network, level)

    # Add custom namespace tags to root element
    _write_extra(root, inventory)

    tree = root.getroottree()

    # The validation has to be done after parsing once again so that the
    # namespaces are correctly assembled.
    if validate is True:
        buf = io.BytesIO()
        tree.write(buf)
        buf.seek(0)
        # This works since ObsPy validate_stationxml gets version from the xml
        validates, errors = validate_stationxml(buf)
        buf.close()
        if validates is False:
            msg = "The created file fails to validate.\n"
            for err in errors:
                msg += "\t%s\n" % err
            raise Exception(msg)

    # Register all namespaces with the tree. This allows for
    # additional namespaces to be added to an inventory that
    # was not created by reading a StationXML file.
    for prefix, ns in nsmap.items():
        if prefix and ns:
            etree.register_namespace(prefix, ns)

    tree.write(file_or_file_object, pretty_print=True, xml_declaration=True,
               encoding="UTF-8")


if __name__ == '__main__':
    main()



  # ****************************************************************************
  #
  # This file is part of the yasmine editing tool.
  #
  # yasmine (Yet Another Station Metadata INformation Editor), a tool to
  # create and edit station metadata information in FDSN stationXML format,
  # is a common development of IRIS and RESIF.
  # Development and addition of new features is shared and agreed between * IRIS and RESIF.
  #
  #
  # Version 1.0 of the software was funded by SAGE, a major facility fully
  # funded by the National Science Foundation (EAR-1261681-SAGE),
  # development done by ISTI and led by IRIS Data Services.
  # Version 2.0 of the software was funded by CNRS and development led by * RESIF.
  #
  # This program is free software; you can redistribute it
  # and/or modify it under the terms of the GNU Lesser General Public
  # License as published by the Free Software Foundation; either
  # version 3 of the License, or (at your option) any later version. *
  # This program is distributed in the hope that it will be
  # useful, but WITHOUT ANY WARRANTY; without even the implied warranty
  # of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  # GNU Lesser General Public License (GNU-LGPL) for more details. *
  # You should have received a copy of the GNU Lesser General Public
  # License along with this software. If not, see
  # <https://www.gnu.org/licenses/>
  #
  #
  # 2019/10/07 : version 2.0.0 initial commit
  #
  # ****************************************************************************/

import os
import sys

import logging
logger = logging.getLogger()

from . import installation_dir

from .libs.libs_log import configure_logger
from .libs.libs_obs import _write_stationxml
from .libs.libs_util import processCmdLine, read_config
from .libs.edit_xml_to_inv import edit_xml_to_inv, plot_responses

def main():
    '''
    usage: yasmine-cli [-h]
                   [--level_network II | --level_station II.* | --level_channel II.ANMO.00.*]
                   [--action [add/delete/update basenode]]
                   [--epoch_station int] [--epoch_channel int] [--field FIELD]
                   [--value VALUE | --from_yml fname.yml] [--infiles] [-o]
                   [-p] [--print_all] [--dont_validate] [--schema_version ver]
                   [--show_fields] [--plot_resp] [--plot_dir path]
                   [--loglevel log level]

    See ../README.md or https://gitlab.isti.com/mhagerty/yasmine-cli
    for full docs

    '''

    fname = 'yasmine-cli'
    config = read_config()
    configure_logger(config, logfile="%s.log" % fname)

    args, scnl_filter = processCmdLine(fname)

    logger.info("[cmd: >%s]" % " ".join(arg for arg in sys.argv))
    logger.info("NET:%s STA:%s LOC:%s CHA:%s" % (scnl_filter.NET, scnl_filter.STA, scnl_filter.LOC, scnl_filter.CHA))
    logger.info("level=[%s] action=[%s]" % (args.level, args.action))

    inv_new, schema_version = edit_xml_to_inv(args, scnl_filter)

    if args.plot_resp:
        plot_responses(inv_new, args.plot_dir)

    else:
        outfile = args.output if args.output else sys.stdout.buffer
        validate = False if args.dont_validate else True

        try:
            #inv_new.write(outfile, format='stationxml', validate=validate)
            # This is a hack to set the output stationxml version to the requested and/or input versions:
            _write_stationxml(inv_new, outfile, validate=validate, schema_version=schema_version)

        except:
            raise

    logger.info("%s: Finished Processing\n" % fname)

    return



# -*- coding: utf-8 -*-
from __future__ import print_function
#from ._version import get_versions

__author__ = 'Mike Hagerty'
__email__ = 'mhagerty@isti.com'
__version__ = '0.0.6'
#__version__ = get_versions()['version']
#del get_versions

import os
def installation_dir():
  return os.path.dirname(os.path.realpath(__file__))

def fdsn_schema_dir():
  return os.path.join(installation_dir(), 'fdsn-schema')

def yml_template_dir():
  return os.path.join(installation_dir(), 'yml')
